/**********************************************************
 * filename:  mpc_tx.c
 * function: encodes simulation data into Crossbow MNAV data
 * packet structure and sends it thru serial port
 *
 * Coder: Rohit Pandita (c) 2006
 * AEM Department, University of Minnesota, Minneapolis
 *
 * changelog:
 * 11-07-08: mpc_tx3: makes use of external serial port block to
 * send data.
 *
 * $Id: mpc_tx3.cpp 259 2011-01-31 22:20:12Z murch $
 ***********************************************************/

#define S_FUNCTION_NAME mpc_tx3
#define S_FUNCTION_LEVEL 2

#include "simstruc.h"

#include <stdio.h>
#include <string.h>

#define IMUONLY_PACKET_LENGTH   51
#define GPS_PACKET_LENGTH	35
#define WITHGPS_PACKET_LENGTH   86

#define	byte	unsigned char
#define word	unsigned short
#define TRUE	1
#define FALSE	0

#define D2R		0.017453292519940
#define R2D		57.29577951308232
#define g2      19.62   	//2*g

#define NIN  8
#define NOUT 1
#define NPAR 0
#define SAMPLE_TIME_0        0.02


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//global variables
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
byte gps_header[3]={'U','U','G'}, packetWithGps_header[3]={'U','U','N'}, imu_header[3]={'U','U','S'};

/*========================================================================*
 *                          Initialization                                *
 *========================================================================*/

static void mdlInitializeSizes(SimStruct *S)
{
    ssSetNumSFcnParams(S, NPAR);
    if(ssGetNumSFcnParams(S) != ssGetSFcnParamsCount(S))   return;
    
    if(!ssSetNumInputPorts(S, NIN)) return;
    
    {
        int_T   i;
        for(i = 0; i < NIN; i++){
            ssSetInputPortDirectFeedThrough(S, i, 1);
            ssSetInputPortDataType(S, i, DYNAMICALLY_TYPED);
        }
        
        ssSetInputPortWidth(S, 0, 3);           //Erik: Sets inport witdth for the structure for each inport based on number of signals muxed in.
        ssSetInputPortWidth(S, 1, 3);
        ssSetInputPortWidth(S, 2, 3);
        ssSetInputPortWidth(S, 3, 3);
        ssSetInputPortWidth(S, 4, 2);
        ssSetInputPortWidth(S, 5, 8);
        ssSetInputPortWidth(S, 6, 1);
        ssSetInputPortWidth(S, 7, 6);
    }
    
    if(!ssSetNumOutputPorts(S,NOUT))    return;  //Erik: set the number of output
    //ports to a nonnegative integer
    
    { int_T i;
    for(i = 0; i < NOUT; i++)
        ssSetOutputPortWidth(S, i, WITHGPS_PACKET_LENGTH);
	ssSetOutputPortDataType( S, 0, SS_UINT8);
    }
    
    ssSetNumSampleTimes(S, 1);
    ssSetNumRWork(S, 0);
    ssSetNumIWork(S, 0);
    ssSetNumPWork(S, 0);
    ssSetNumModes(S, 0);
    ssSetNumNonsampledZCs(S, 0);
    
}


static void mdlInitializeSampleTimes(SimStruct *S)
{
    ssSetSampleTime(S, 0, INHERITED_SAMPLE_TIME);
    ssSetOffsetTime(S, 0, 0.0);
}

#undef MDL_INITIALIZE_CONDITIONS   /* Change to #undef to remove function */
#if defined(MDL_INITIALIZE_CONDITIONS)

static void mdlInitializeConditions(SimStruct *S)
{
    
}
#endif /* MDL_INITIALIZE_CONDITIONS */



/*========================================================================*
 *                     Simulation Start Time                              *
 *========================================================================*/

#define MDL_START  /* Change to #undef to remove function */
#if defined(MDL_START)

static void mdlStart(SimStruct *S)
{
}
#endif /*  MDL_START */

// functions
unsigned short do_chksum(byte* buffer, int packet_len)
{
    word     	 i=0,rcvchecksum=0;
    unsigned short sum=0;
    
    for(i=2;i<packet_len-2;i++) sum += buffer[i];
    
    return  sum;
}

short bigEnd (short s)
{
    unsigned char c1, c2;
    
    c1 = s & 0xFF;
    c2 = (s >> 8) & 0xFF;
    
    return ((c1 << 8) + c2);
}

/*========================================================================*
 *                              Output                                    *
 *========================================================================*/

static void mdlOutputs(SimStruct *S, int_T tid)
{
    // local variables
    byte gps_packet[GPS_PACKET_LENGTH]={'U','U','G',};
	byte mnav_packet[WITHGPS_PACKET_LENGTH]={0,};
    word imu_data[14];
	word ppm_data[9];
	word gps_data[15];
    
    float ax,ay,az,p,q,r,hx,hy,hz,tx,ty,tz,ps,pt,vn,ve,vd,lon,lat,alt;
    //word ch0,ch1,ch2,ch3,ch4,ch5,ch6,ch7;
    int gpsFlag, i;
        
    uint8_T *y0 = (uint8_T *)ssGetOutputPortSignal(S,0);

    // imudata
    ax = *ssGetInputPortRealSignalPtrs(S, 0)[0];
    ay = *ssGetInputPortRealSignalPtrs(S, 0)[1];
    az = *ssGetInputPortRealSignalPtrs(S, 0)[2];
    p  = *ssGetInputPortRealSignalPtrs(S, 1)[0];
    q  = *ssGetInputPortRealSignalPtrs(S, 1)[1];
    r  = *ssGetInputPortRealSignalPtrs(S, 1)[2];
    hx = *ssGetInputPortRealSignalPtrs(S, 2)[0];
    hy = *ssGetInputPortRealSignalPtrs(S, 2)[1];
    hz = *ssGetInputPortRealSignalPtrs(S, 2)[2];
    tx = *ssGetInputPortRealSignalPtrs(S, 3)[0];
    ty = *ssGetInputPortRealSignalPtrs(S, 3)[1];
    tz = *ssGetInputPortRealSignalPtrs(S, 3)[2];
    ps = *ssGetInputPortRealSignalPtrs(S, 4)[0];
    pt = *ssGetInputPortRealSignalPtrs(S, 4)[1];
        
    // PPM not used anymore; AMM 1/12/11
    /*
    // Servo data
    ch0 = (unsigned short)*ssGetInputPortRealSignalPtrs(S, 5)[0];
    ch1 = (unsigned short)*ssGetInputPortRealSignalPtrs(S, 5)[1];
    ch2 = (unsigned short)*ssGetInputPortRealSignalPtrs(S, 5)[2];
    ch3 = (unsigned short)*ssGetInputPortRealSignalPtrs(S, 5)[3];
    ch4 = (unsigned short)*ssGetInputPortRealSignalPtrs(S, 5)[4];
    ch5 = (unsigned short)*ssGetInputPortRealSignalPtrs(S, 5)[5];
    ch6 = (unsigned short)*ssGetInputPortRealSignalPtrs(S, 5)[6];
    ch7 = (unsigned short)*ssGetInputPortRealSignalPtrs(S, 5)[7];
    */
    
    // GPS data
    gpsFlag = *ssGetInputPortRealSignalPtrs(S, 6)[0];
    vn = *ssGetInputPortRealSignalPtrs(S, 7)[0];
    ve = *ssGetInputPortRealSignalPtrs(S, 7)[1];
    vd = *ssGetInputPortRealSignalPtrs(S, 7)[2];
    lon= *ssGetInputPortRealSignalPtrs(S, 7)[3];
    lat= *ssGetInputPortRealSignalPtrs(S, 7)[4];
    alt= *ssGetInputPortRealSignalPtrs(S, 7)[5];
    
    
    // construct sensor data packet
    
    imu_data[0] = bigEnd((word)( ax / g2 * 0x7FFF ));	//|accel| <=  19.62 m/s^2
    imu_data[1] = bigEnd((word)( ay / g2 * 0x7FFF ));
    imu_data[2] = bigEnd((word)( az / g2 * 0x7FFF ));
    
    imu_data[3] = bigEnd((word)( p*R2D / 200.0 * 0x7FFF ));	//|RR| <= 200 deg/s
    imu_data[4] = bigEnd((word)( q*R2D / 200.0 * 0x7FFF ));
    imu_data[5] = bigEnd((word)( r*R2D / 200.0 * 0x7FFF ));
    
    imu_data[6] = bigEnd((word)( hx*1e-05 / 2.0 *0x7FFF ));		// <= 2 Gauss
    imu_data[7] = bigEnd((word)( hy*1e-05 / 2.0 *0x7FFF ));
    imu_data[8] = bigEnd((word)( hz*1e-05 / 2.0 *0x7FFF ));
    
    //    imu_data[9] = (word)( tx / 350.0 * 0x7FFF );	// |T| <=  350Kelvin        //Original
    //    imu_data[10]= (word)( ty / 350.0 * 0x7FFF );
    //    imu_data[11]= (word)( tz / 350.0 * 0x7FFF );
    
    imu_data[9] = bigEnd((word)( (tx-273.15) / 200.0 * 0x7FFF ));	// |T| <=  350Kelvin        //Erik's Switch
    imu_data[10]= bigEnd((word)( (ty-273.15) / 200.0 * 0x7FFF ));	//This comes out correct
    imu_data[11]= bigEnd((word)( (tz-273.15) / 200.0 * 0x7FFF ));   //See MNAV XBOW user Manual
    
    imu_data[12]= bigEnd((word)( ps / 10000.0 * 0x7FFF ));	// Pressure Alt.(m) = 10000 m
    imu_data[13]= bigEnd((word)( pt / 80.0 * 0x7FFF ));		// Equivalent Airspeed(EAS) = 80 m/s
    
    // PPM not used anymore; AMM 1/12/11
    //ppm_data[0] = {'F','F'};
    ppm_data[0] = 'F';   ppm_data[0] = ppm_data[0] << 8 | 'F';
    /*
    ppm_data[1] = bigEnd(ch0);
    ppm_data[2] = bigEnd(ch1);
    ppm_data[3] = bigEnd(ch2);
    ppm_data[4] = bigEnd(ch3);
    ppm_data[5] = bigEnd(ch4);
    ppm_data[6] = bigEnd(ch5);
    ppm_data[7] = bigEnd(ch6);
    ppm_data[8] = bigEnd(ch7);
    */
    
    /*
	// gps data
    *(int *)&gps_data[0]=(int)(vn);
    *(unsigned int *)&gps_data[2]=(ve);
    *(int *)&gps_data[4]=(int)(vd);
    *(int *)&gps_data[6]=(int)(lon*1e07);
    *(int *)&gps_data[8]=(int)(lat*1e07);
    *(int *)&gps_data[10]=(int)(alt*1e03);
    */
	
	// gps data
    *(unsigned int *)&gps_data[0] =(vn);
    *(unsigned int *)&gps_data[2] =(ve);
    *(unsigned int *)&gps_data[4] =(vd);
    *(unsigned int *)&gps_data[6] =(lon*1e07);
    *(unsigned int *)&gps_data[8] =(lat*1e07);
    *(unsigned int *)&gps_data[10]=(alt*1e03);
    
    // unused
    gps_data[12] = 0;
    gps_data[13] = 0;
    gps_data[14] = 0;
    
    //memcpy (&gps_packet[0], gps_header, 3);
    memcpy (&gps_packet[3], gps_data, 30);
    *(word *)&gps_packet[33] = bigEnd(do_chksum (gps_packet, GPS_PACKET_LENGTH));
    
       
    // assemble the output data packet according to xbow micronav structure (ref: MNAV1000CA_user_manual.pdf)
    if (gpsFlag == FALSE)
    {
        memcpy(&mnav_packet, imu_header,3);
        memcpy(&mnav_packet[3], imu_data, 28);
        memcpy(&mnav_packet[31], ppm_data, 18);
        *(word *)&mnav_packet[49] = bigEnd(do_chksum (mnav_packet, IMUONLY_PACKET_LENGTH));

    for (i=0; i<IMUONLY_PACKET_LENGTH; i++)
	{
		y0[i] = mnav_packet[i]; 
	};
	
    }
    
    else if (gpsFlag == TRUE)
    {
        memcpy(&mnav_packet, packetWithGps_header,3);
        memcpy(&mnav_packet[3], imu_data, 28);
        memcpy(&mnav_packet[31], gps_packet, 35);
        memcpy(&mnav_packet[66], ppm_data, 18);
        *(word *)&mnav_packet[84] = bigEnd(do_chksum (mnav_packet, WITHGPS_PACKET_LENGTH));        

    for (i=0; i<WITHGPS_PACKET_LENGTH; i++)
	{
		y0[i] = mnav_packet[i]; 
	};
	
    }


}


/*========================================================================*
 *                              Termination                               *
 *========================================================================*/

static void mdlTerminate(SimStruct *S)
{
    
}


#ifdef  MATLAB_MEX_FILE    /* Is this file being compiled as a MEX-file? */
#include "simulink.c"      /* MEX-file interface mechanism */
#else
#include "cg_sfun.h"       /* Code generation registration function */
#endif
