/******************************************************************************
 * FILE: L_datalogger_ramfs.c
 * DESCRIPTION:
 * Logs binary flight data onboard MPC5200 flash memory on in RAMFS as per user
 * configuration.
 *
 * AUTHOR: Rohit Pandita
 *
 * University of Minnesota 
 * Aerospace Engineering and Mechanics 
 *  2011 Regents of the University of Minnesota. All rights reserved.
 *
 * $Id: L_datalogger_ramfs.c 253 2011-01-20 22:01:35Z murch $
 ******************************************************************************/
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>


#include <cyg/io/flash.h>
#include <cyg/fileio/fileio.h>
#include <pkgconf/io_fileio.h>
#include <pkgconf/io_flash.h>
#include <pkgconf/fs_jffs2.h>	// Address of JFFS2

#include <sys/socket.h>
#include <network.h>
#include <tftp_support.h>		// TFTP support
#include <netdb.h>
#include <arpa/inet.h>

#define putfilename 	  		"/tftpboot/flightdata.bin"

#ifdef DATALOG_RAMFS
#include <pkgconf/fs_ram.h>
#define FLIGHTDATA_FILE	  		"flightdata.ram"
#define getfilename		  	    "/flightdata.ram"
#endif

#ifdef DATALOG_FLASH
#define FLIGHTDATA_FILE	  		"flightdata.bin"
#define getfilename		  	    "/flightdata.bin"
#endif


int createfile( char *name );
int closefile ( int fd );
void checkcwd( const char *cwd );
void makeDatalogPacket( struct imu *imupkt, struct gps *gpspkt, struct nav *navpkt,
		struct flightdat2 *outpkt );
int tftpTransfer(char *outfile, char *txbuf, int len);
char *tftp_error(int err);
void readfile (int fd, char *buf, int size);
int filetransferToGCS( char *readfilename, char *writefilename );
int umountFilesystem();
void endian_swap( unsigned char *buf, int index, int count );


#define stringify2(_x_) #_x_
#define stringify(_x_) stringify2(_x_)

#if defined(CYGDAT_IO_FLASH_BLOCK_DEVICE_NAME_1)
# define JFFS2_TEST_DEV CYGDAT_IO_FLASH_BLOCK_DEVICE_NAME_1
#elif defined(CYGFUN_IO_FLASH_BLOCK_FROM_FIS)
# define JFFS2_TEST_DEV "/dev/flash/fis/jffs2test"
#else
// fall back to using a user set area in the first device (only)
# define JFFS2_TEST_DEV "/dev/flash/0/" stringify(CYGNUM_FS_JFFS2_TEST_OFFSET) \
		"," stringify(CYGNUM_FS_JFFS2_TEST_LENGTH)
#endif
/*
// diagnostic error messages
#define SHOW_RESULT( _fn, _res ) \
		fprintf(stderr, "<FAIL>: " #_fn "() returned %ld %s\n",  \
				(unsigned long)_res, _res<0?strerror(errno):"");
 */
#define SHOW_RESULT( _fn, _res ) \
		diag_printf("<FAIL>: " #_fn "() returned %ld %s\n",  \
				(unsigned long)_res, _res<0?strerror(errno):"");

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//globals
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++



void *datalogger(void *thread_id)
{
	double 		toc;
	static double 	tic;
	int wrote, rc, i;


	/*
	info =  mallinfo();		// Free memory in the heap at the beginning of datalogging operation
	diag_printf("arenasize %d, freeblocks %d, totalallocated %d, totalfree %d, maxfree %d\n",\
	info.arena, info.ordblks, info.uordblks, info.fordblks, info.maxfree);
	 */

	pthread_mutex_lock(&mutex_sPort0);
	sprintf(&msg[3],"DATALOG thread created");
	write(sPort0, msg, sizeof(msg));
	pthread_mutex_unlock(&mutex_sPort0);


	// Create and open a file for dumping flight data to
	fileHandle1 = createfile (FLIGHTDATA_FILE);
	if (fileHandle1 > 0) {
		pthread_mutex_lock(&mutex_sPort0);
		sprintf(&msg[3],"<DNLINK>: Logging data");
		write(sPort0, msg, sizeof(msg));
		pthread_mutex_unlock(&mutex_sPort0);
	}
	else {
		pthread_mutex_lock(&mutex_sPort0);
		sprintf(&msg[3],"<DNLINK>: NO FILEHANDLE. returned %d",fileHandle1);
		write(sPort0, msg, sizeof(msg));
		pthread_mutex_unlock(&mutex_sPort0);
	}


	while( ofpMode != terminate )
	{
		pthread_mutex_lock (&mutex_imu);
		rc = pthread_cond_wait (&trigger_datalogger, &mutex_imu);

		if (rc == 0) {

			// compute thread jitter
			toc = get_Time();
			threadstat[5].jitter_nsec = ( (toc-tic) - 1/(double)DATALOGGER_Hz ) * 1E09;
			tic = toc;

			pthread_mutex_lock(&mutex_fdat);
			pthread_mutex_lock(&mutex_gps);

			// thread statistics
			for( i=0; i<6; i++)	fltdatpacket.threadJitter[i] = threadstat[i].jitter_nsec;
			makeDatalogPacket( &imupacket, &gpspacket, &navpacket, &fltdatpacket );

			pthread_mutex_unlock(&mutex_gps);
			pthread_mutex_unlock(&mutex_fdat);
			pthread_mutex_unlock(&mutex_imu);

			wrote = write (fileHandle1, &fltdatpacket, sizeof(fltdatpacket));

			if ( wrote < 0 ) {
				pthread_mutex_lock(&mutex_sPort0);
				//fprintf(,"%s",strerror(errno));
				sprintf(&msg[3],"<DNLINK>:%s",strerror(errno));
				write(sPort0, msg, sizeof(msg));
				pthread_mutex_unlock(&mutex_sPort0);
			}
		}
	}	/* end while */

	closefile( fileHandle1 );
	pthread_exit(NULL);
}


//------------------------------- functions -----------------------------------

void makeDatalogPacket( struct imu *imupkt, struct gps *gpspkt, struct nav *navpkt,
		struct flightdat2 *outpkt )
{
	int i;

	// body angular rates
	outpkt->p = (float)(imupkt->p);
	outpkt->q = (float)(imupkt->q);
	outpkt->r = (float)(imupkt->r);

	// accelerometer
	outpkt->ax = (float)(imupkt->ax);
	outpkt->ay = (float)(imupkt->ay);
	outpkt->az = (float)(imupkt->az);

	// magnetometers
	outpkt->hx = (float)(imupkt->hx);
	outpkt->hy = (float)(imupkt->hy);
	outpkt->hz = (float)(imupkt->hz);

	// Euler attitude estimates
	outpkt->psi = (float)(imupkt->psi);
	outpkt->the = (float)(imupkt->the);
	outpkt->phi = (float)(imupkt->phi);

	// altitude (AGL) , Indicated airspeed (IAS), and pressures
	outpkt->h   = (float)(imupkt->h);
	outpkt->ias = (float)(imupkt->ias);
	outpkt->Ps  = (float)(imupkt->Ps);
	outpkt->Pd  = (float)(imupkt->Pd);	

	// GPS
	outpkt->lat = (float)(gpspkt->lat);
	outpkt->lon = (float)(gpspkt->lon);
	outpkt->alt = (float)(gpspkt->alt);
	outpkt->vn = (float)(gpspkt->vn);
	outpkt->ve = (float)(gpspkt->ve);
	outpkt->vd = (float)(gpspkt->vd);

	// NAV
	outpkt->navlat = (float)(navpkt->lat);
	outpkt->navlon = (float)(navpkt->lon);
	outpkt->navalt = (float)(navpkt->alt);
	outpkt->navvn = (float)(navpkt->vn);
	outpkt->navve = (float)(navpkt->ve);
	outpkt->navvd = (float)(navpkt->vd);

	// status msg
	outpkt->imu_status = imupkt->err_type;
	outpkt->gps_status = gpspkt->err_type;

	// imu packet time
	//outpkt->time = (float)(imupkt->time);
	outpkt->time = get_Time();

	/*
	 **note** : control thread specific variables are \
	populated into the struct by control thread
	 */
#ifdef DATALOG_LSB
	endian_swap( (byte *)&(outpkt->p), 0, 4 );
	endian_swap( (byte *)&(outpkt->q), 0, 4 );
	endian_swap( (byte *)&(outpkt->r), 0, 4 );

	endian_swap( (byte *)&(outpkt->ax), 0, 4 );
	endian_swap( (byte *)&(outpkt->ay), 0, 4 );
	endian_swap( (byte *)&(outpkt->az), 0, 4 );

	endian_swap( (byte *)&(outpkt->hx), 0, 4 );
	endian_swap( (byte *)&(outpkt->hy), 0, 4 );
	endian_swap( (byte *)&(outpkt->hz), 0, 4 );

	endian_swap( (byte *)&(outpkt->psi), 0, 4 );
	endian_swap( (byte *)&(outpkt->the), 0, 4 );
	endian_swap( (byte *)&(outpkt->phi), 0, 4 );

	endian_swap( (byte *)&(outpkt->h), 0, 4 );
	endian_swap( (byte *)&(outpkt->ias), 0, 4 );
	endian_swap( (byte *)&(outpkt->Ps), 0, 4 );
	endian_swap( (byte *)&(outpkt->Pd), 0, 4 );	

	endian_swap( (byte *)&(outpkt->lat), 0, 4 );
	endian_swap( (byte *)&(outpkt->lon), 0, 4 );
	endian_swap( (byte *)&(outpkt->alt), 0, 4 );
	endian_swap( (byte *)&(outpkt->vn), 0, 4 );
	endian_swap( (byte *)&(outpkt->ve), 0, 4 );
	endian_swap( (byte *)&(outpkt->vd), 0, 4 );

	endian_swap( (byte *)&(outpkt->navlat), 0, 4 );
	endian_swap( (byte *)&(outpkt->navlon), 0, 4 );
	endian_swap( (byte *)&(outpkt->navalt), 0, 4 );
	endian_swap( (byte *)&(outpkt->navvn), 0, 4 );
	endian_swap( (byte *)&(outpkt->navve), 0, 4 );
	endian_swap( (byte *)&(outpkt->navvd), 0, 4 );

	endian_swap( (byte *)&(outpkt->imu_status), 0, sizeof(outpkt->imu_status) );
	endian_swap( (byte *)&(outpkt->gps_status), 0, sizeof(outpkt->gps_status) );
	endian_swap( (byte *)&(outpkt->time), 0, 4 );

	for( i=0; i<6; i++)
			endian_swap( (byte *)&(outpkt->threadJitter[i]), 0, sizeof(int) );

#endif
}

int createfile( char *name )
{
	int fd;
	int err;

#ifdef DATALOG_FLASH
	err = mount( JFFS2_TEST_DEV, "/", "jffs2" );
	if( err < 0 ) {
		SHOW_RESULT( mount, err );
	}
	else
	{
		pthread_mutex_lock(&mutex_sPort0);
		sprintf(&msg[3],"<DNLINK>: %s mounted successfully",JFFS2_TEST_DEV);
		write(sPort0, msg, sizeof(msg));
		pthread_mutex_unlock(&mutex_sPort0);
	}
#endif

#ifdef DATALOG_RAMFS
	err = mount( "", "/", "ramfs" );
	if( err < 0 ) {
		SHOW_RESULT( mount, err );
	}
	else
	{
		pthread_mutex_lock(&mutex_sPort0);
		sprintf(&msg[3],"<DNLINK>: RAMFS mounted successfully");
		write(sPort0, msg, sizeof(msg));
		pthread_mutex_unlock(&mutex_sPort0);
	}
#endif


	err = chdir( "/" );
	if( err < 0 ) SHOW_RESULT( chdir, err );

	checkcwd( "/" );

	// if a datafile already exists, delete it
	if ( access( name, F_OK ) == 0 )
		unlink(name);

	fd = open( name, O_RDWR | O_CREAT );
	if( fd < 0 ) {
		//SHOW_RESULT( open, fd );
		pthread_mutex_lock(&mutex_sPort0);
		sprintf(&msg[3],"<DNLINK>: createfile() error, %s", strerror(fd));
		write(sPort0, msg, sizeof(msg));
		pthread_mutex_unlock(&mutex_sPort0);
	}

	return fd;
}

int closefile (int fd)
{
	int err;

	err = close( fd );
	if( err < 0 ) SHOW_RESULT( close, err );


	return err;
}

int umountFilesystem()
{
	int err;

	// delete datalog file from filesystem
	err = unlink(FLIGHTDATA_FILE);
	if( err < 0 ){
		diag_printf("Deleting file %s in RAMFS failed.\n",FLIGHTDATA_FILE);
	}

	// unmount FS
	err = umount("/");

	if( err < 0 ){
		diag_printf("\n Unmounting FileSystem failed.\n");
		//		SHOW_RESULT( umount, err );
		return err;
	}
	else
		return 0;
}

int filetransferToGCS( char *readfilename, char *writefilename )
{

	int err, fd, filesize=0;
	struct stat filestatus;

	err = access( readfilename, F_OK );
	if( err < 0 && errno != EACCES )
		SHOW_RESULT( access, err );

	// open file to transfer
	fd = open( readfilename, O_RDONLY );
	if( fd < 0 )
		SHOW_RESULT( open, fd );

	if( fstat( fd, &filestatus ) < 0 ){};

	info =  mallinfo();		// Free memory in the heap at the end of datalogging operation
	fprintf(stderr,"\n Tot. allocated memory:%d, Max. free memory: %d\n", info.uordblks, info.fordblks);

	filesize = (size_t)(filestatus.st_size);
	/*
	if( info.maxfree < filesize )
		fprintf(stderr,"\n Insufficient memory for filetransfer !!\n");

	// allocated memory to store file contents in buffer
	buf = (char *)malloc(filesize);

	if (buf == NULL)
		fprintf(stderr,"\n malloc() failed");
	else
		fprintf(stderr,"\n malloc() allocated %d bytes to buf for filetransfer (TFTP).\n", (int)filesize);
	 */
	// read file contents to a temp buffer
	readfile ( fd, buf, filesize );
	closefile( fd );

	tftpTransfer(writefilename, buf, filesize);

	//free(buf);
	return filesize;
}

void checkcwd( const char *cwd )
{
	static char cwdbuf[PATH_MAX];
	char *ret;

	ret = getcwd( cwdbuf, sizeof(cwdbuf));
	if( ret == NULL ) SHOW_RESULT( getcwd, ret );

	if( strcmp( cwdbuf, cwd ) != 0 )
	{
		//		diag_printf( "cwdbuf %s cwd %s\n",cwdbuf, cwd );
	}
}

void readfile (int fd, char *buf, int size)
{
	int done;
	done = read( fd, &buf[0], size );
	if( done < 0 ) SHOW_RESULT( read, done );
}

int tftpTransfer(char *outfile, char *txbuf, int len)
{
	int res, err;
	struct sockaddr_in host;

	//init_all_network_interfaces();


	build_bootp_record(&eth0_bootp_data,
			eth0_name,
			"192.168.3.11",		// eth0 ip_addr
			"255.255.255.0",	// netmask
			"192.168.3.255",		// broadcast
			"192.168.3.1",		// gateway
			"192.168.3.10");	// server ip_addr


	struct bootp *bp = &eth0_bootp_data;

	//show_bootp(eth0_name, bp);

	// initialize 'eth0' network interface
	if (!init_net(eth0_name, bp)) {
		diag_printf("Network initialization failed for eth0\n");
	}

	sleep(2);	// wait for connection to be established

	memset((char *)&host, 0, sizeof(host));
	host.sin_len = sizeof(host);
	host.sin_family = AF_INET;
	host.sin_addr = bp->bp_siaddr;
	host.sin_port = 0;

	fprintf(stderr,"Initiating TFTP filetransfer to %s to Host:%16s, length %d\n",
			outfile, inet_ntoa(host.sin_addr), len);


	res = tftp_put( outfile, &host, txbuf, len, TFTP_OCTET, &err);

	if (res < 0) {
		SHOW_RESULT (tftp_put, err);
		fprintf(stderr,"%s",tftp_error(err));
		return(res);
	}
	else {
		//fprintf(stderr,"\n Transfer successful. %d bytes transferred", res);
		diag_printf("\n Transfer successful. %d bytes transferred", res);
		return(res);
	}

}

char *tftp_error(int err)
{
	char *errmsg = "Unknown error";

	switch (err) {
	case TFTP_ENOTFOUND:
		return "file not found";
	case TFTP_EACCESS:
		return "access violation";
	case TFTP_ENOSPACE:
		return "disk full or allocation exceeded";
	case TFTP_EBADOP:
		return "illegal TFTP operation";
	case TFTP_EBADID:
		return "unknown transfer ID";
	case TFTP_EEXISTS:
		return "file already exists";
	case TFTP_ENOUSER:
		return "no such user";
	case TFTP_TIMEOUT:
		return "operation timed out";
	case TFTP_NETERR:
		return "some sort of network error";
	case TFTP_INVALID:
		return "invalid parameter";
	case TFTP_PROTOCOL:
		return "protocol violation";
	case TFTP_TOOLARGE:
		return "file is larger than buffer";
	}
	return errmsg;
}

// swap big/little endian bytes
void endian_swap( unsigned char *buf, int index, int count )
{
	int i;
	unsigned char tmp;

	for ( i = 0; i < count / 2; ++i ) {
		tmp = buf[index+i];
		buf[index+i] = buf[index+count-i-1];
		buf[index+count-i-1] = tmp;
	}
}
